#! /usr/local/bin/python
# -*- coding: utf-8 -*-
##
## chat_sever.py: ein simpler Chat-Server
## Version 0.13, 2002-10-02
## Schule am Ried (www.Schule-am-Ried.de)
## W. Spiegel (walter.spiegel@web.de)
## Magic Key:
## Windows ==>  Strg + Untbr/Pause bzw. Ctrl + Break
## Linux   ==>  Strg + C   bzw. Ctrl + C
## To Do:
##      - Mitprotokollieren in Datei
##      - tkinter-Oberfläche
##      - Benutzer rauskicken
##      - Einstellungen
##
# stark verändert nach M. Lutz: Programming Python
# ch. 10, S. 547ff (Broadcast + locking hinzugefügt)
# with a little Help from J. Bishop: Java lernen, S. 524ff
#########################################################
# Server side: open a socket on a port, listen for
# a message from a client, and send a broadcast; 
# echos lines until 'ENDE' when client closes socket;
# spawns a thread to handle each client connection;
# threads share global memory space with main thread;
#########################################################

import thread, time, string, os
from socket import *        # get socket constructor and constants
myHost = ''                 # server machine, '' means local host
myPort = 8888              # listen on a non-reserved port number
nick_flag = 0               # 0: Adresse, 1: Nick-Name
nicki={}                    # enthält Adresse und Nick-Name
conn={}                     # enthält Adresse und connection-Parameter   
conn_lock = thread.allocate_lock()

sockobj = socket(AF_INET, SOCK_STREAM)  # make a TCP socket object
sockobj.bind((myHost, myPort))          # bind it to server port number
sockobj.listen(5)                       # allow upto 5 pending connects

if len(myHost) > 0:
    print 'Server eingerichtet an Adresse',myHost,'auf Port',str(myPort)
else:
    print 'Server eingerichtet auf Port',str(myPort)

if os.name == 'posix': 
    print 'Magic Key: <Strg> + <C> [<Ctrl> + <C>]'
else: 
    print 'Magic Key: <Strg> + <Untbr> [<Ctrl> + <Break>]'

#######################    
def now(): 
    return time.ctime(time.time())      # current time on the server
#######################    
def broadcast(msg, name):
    global conn
    conn_lock.acquire()
    chatter = conn.keys()
    for address in chatter:
        conn[address].send(name +': '+ msg)
    conn_lock.release()
#######################    
def handleClient(connection, address):  # in spawned thread: reply
    global nicki,conn
    
    connection.send('Verbunden! Beenden mit "ENDE", dein Nickname???')
    data = connection.recv(1024)
    if data:  ## doppelte Nicknamen verhindern!       
        while (data and data in nicki.values() or string.lower(data) in nicki.values() or string.upper(data) in nicki.values()):
            connection.send('Nickname "' + data + '" schon vergeben!!!! Denk dir einen neuen Namen aus!!')
            data = connection.recv(1024)
            
    if data:        
        nicki[address]=data        
        conn_lock.acquire()        
        conn[address]=connection
        conn_lock.release()        
        broadcast(data + ' im Chatraum! ','*** SERVER ***')    

    while 1:                        # read, write a client socket
        data = connection.recv(1024)
        if not data:
            break                   # keine Daten, das wars!
        if nick_flag:
            print "Daten von '" + nicki.get(address) + "' erhalten: " + data
        else:
            print "Daten von " + str(address) + " erhalten: " + data            

        if data != 'ENDE':
            broadcast (data,nicki.get(address) )
        else:  # data == 'ENDE', beenden!
            connection.send('OAIC') # magic string ciao rückwärts!
            conn_lock.acquire()
            del conn[address]
            conn_lock.release()            
            broadcast('Ciao ' + nicki.get(address) ,'*** SERVER ***')
            del nicki[address]  
            break
        
    print 'Beende Verbindung mit', address
    connection.close()         

#######################    
def dispatcher():                                # listen until process killd
    while 1:                                     # wait for next connection,
        connection, address = sockobj.accept()   # pass to thread for service
        print 'Server verbunden mit', address,
        print ',', now() 
        thread.start_new(handleClient, (connection,address))
################# Hauptprogramm!
dispatcher()
